// D import file generated from 'tango/time/chrono/Gregorian.d'
module tango.time.chrono.Gregorian;
private import tango.time.chrono.Calendar;

private import tango.core.Exception;

class Gregorian : Calendar
{
    alias Calendar.toTime toTime;
    public static __gshared Gregorian generic;


    enum Type 
{
Localized = 1,
USEnglish = 2,
MiddleEastFrench = 9,
Arabic = 10,
TransliteratedEnglish = 11,
TransliteratedFrench = 12,
}
    private Type type_;

    enum 
{
AD_ERA = 1,
BC_ERA = 2,
MAX_YEAR = 9999,
}
    private static enum uint[] DaysToMonthCommon = [0,31,59,90,120,151,181,212,243,273,304,334,365];


    private static enum uint[] DaysToMonthLeap = [0,31,60,91,121,152,182,213,244,274,305,335,366];


    shared static this();
    this(Type type = Type.Localized)
{
type_ = type;
}
    const override Time toTime(uint year, uint month, uint day, uint hour, uint minute, uint second, uint millisecond, uint era)
{
return Time(getDateTicks(year,month,day,era) + getTimeTicks(hour,minute,second)) + TimeSpan.fromMillis(millisecond);
}

    const override DayOfWeek getDayOfWeek(const(Time) time)
{
auto ticks = time.ticks;
int offset = 1;
if (ticks < 0)
{
++ticks;
offset = 0;
}
auto dow = cast(int)((ticks / TimeSpan.TicksPerDay + offset) % 7);
if (dow < 0)
dow += 7;
return cast(DayOfWeek)dow;
}

    const override uint getDayOfMonth(const(Time) time)
{
return extractPart(time.ticks,DatePart.Day);
}

    const override uint getDayOfYear(const(Time) time)
{
return extractPart(time.ticks,DatePart.DayOfYear);
}

    const override uint getMonth(const(Time) time)
{
return extractPart(time.ticks,DatePart.Month);
}

    const override uint getYear(const(Time) time)
{
return extractPart(time.ticks,DatePart.Year);
}

    const override uint getEra(const(Time) time)
{
if (time < time.epoch)
return BC_ERA;
else
return AD_ERA;
}

    const override uint getDaysInMonth(uint year, uint month, uint era)
{
if (month < 1 || month > 12)
argumentError("months out of range");
auto monthDays = isLeapYear(year,era) ? DaysToMonthLeap : DaysToMonthCommon;
return monthDays[month] - monthDays[month - 1];
}

    const override uint getDaysInYear(uint year, uint era)
{
return isLeapYear(year,era) ? 366 : 365;
}

    const override uint getMonthsInYear(uint year, uint era)
{
return 12;
}

    const override bool isLeapYear(uint year, uint era)
{
return staticIsLeapYear(year,era);
}

    const @property Type calendarType()
{
return type_;
}

    const override @property uint[] eras()
{
uint[] tmp = [AD_ERA,BC_ERA];
return tmp.dup;
}

    const override @property uint id()
{
return cast(int)type_;
}

    const override void split(const(Time) time, ref uint year, ref uint month, ref uint day, ref uint doy, ref uint dow, ref uint era)
{
splitDate(time.ticks,year,month,day,doy,era);
dow = getDayOfWeek(time);
}

    const override Time addMonths(const(Time) t, int nMonths, bool truncateDay = false)
{
auto date = toDate(t);
nMonths += date.month - 1;
int nYears = nMonths / 12;
nMonths %= 12;
if (nMonths < 0)
{
nYears--;
nMonths += 12;
}
int realYear = date.year;
if (date.era == BC_ERA)
realYear = -realYear + 1;
realYear += nYears;
if (realYear < 1)
{
date.year = -realYear + 1;
date.era = BC_ERA;
}
else
{
date.year = realYear;
date.era = AD_ERA;
}
date.month = nMonths + 1;
if (truncateDay)
{
uint maxday = getDaysInMonth(date.year,date.month,date.era);
if (date.day > maxday)
date.day = maxday;
}
auto tod = t.ticks % TimeSpan.TicksPerDay;
if (tod < 0)
tod += TimeSpan.TicksPerDay;
return toTime(date) + TimeSpan(tod);
}

    const override Time addYears(const(Time) t, int nYears)
{
return addMonths(t,nYears * 12);
}

    package static void splitDate(long ticks, ref uint year, ref uint month, ref uint day, ref uint dayOfYear, ref uint era);


    package static uint extractPart(long ticks, DatePart part)
{
uint year,month,day,dayOfYear,era;
splitDate(ticks,year,month,day,dayOfYear,era);
if (part is DatePart.Year)
return year;
if (part is DatePart.Month)
return month;
if (part is DatePart.DayOfYear)
return dayOfYear;
return day;
}


    package static long getDateTicks(uint year, uint month, uint day, uint era)
{
if (day < 1 || day > generic.getDaysInMonth(year,month,era))
argumentError("days out of range");
auto monthDays = staticIsLeapYear(year,era) ? DaysToMonthLeap : DaysToMonthCommon;
if (era == BC_ERA)
{
year += 2;
return -cast(long)((year - 3) * 365 + year / 4 - year / 100 + year / 400 + monthDays[12] - (monthDays[month - 1] + day - 1)) * TimeSpan.TicksPerDay;
}
else
{
year--;
return (year * 365 + year / 4 - year / 100 + year / 400 + monthDays[month - 1] + day - 1) * TimeSpan.TicksPerDay;
}
}


    package static bool staticIsLeapYear(uint year, uint era)
{
if (year < 1)
argumentError("year cannot be 0");
if (era == BC_ERA)
{
if (year == 1)
return true;
return staticIsLeapYear(year - 1,AD_ERA);
}
if (era == AD_ERA || era == CURRENT_ERA)
return year % 4 == 0 && (year % 100 != 0 || year % 400 == 0);
return false;
}


    package static void argumentError(const(char[]) str);


}
debug (Gregorian)
{
    import tango.io.Stdout;
    void output(Time t)
{
Date d = Gregorian.generic.toDate(t);
TimeOfDay tod = t.time;
Stdout.format("{}/{}/{:d4} {} {}:{:d2}:{:d2}.{:d3} dow:{}",d.month,d.day,d.year,d.era == Gregorian.AD_ERA ? "AD" : "BC",tod.hours,tod.minutes,tod.seconds,tod.millis,d.dow).newline;
}
    void main()
{
Time t = Time(365 * TimeSpan.TicksPerDay);
output(t);
{
for (int i = 0;
 i < 366 + 365; i++)
{
{
t -= TimeSpan.fromDays(1);
output(t);
}
}
}
}
}
debug (UnitTest)
{
    }
